# glam: Galaxy Line Analyzer with MCMC
#Copyright (c) 2024 Livia Vallini

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
import numpy as np

"""
    this module contains the analytical eq.s for the cooling rate of various ions
    outpus:
      cooling rates in erg/s/cm-3
    inputs:
      T, the temperature in K
"""

def compute_lambda_CII_h(T):
    
    """
    Maxwellian-averaged collision rates with neutrals
    using expression from Goldsmith et al. 2012
    see Appendix A, Ferrara et al. 2019
    """

    factor = (1.84e-4*(T)**0.64)/2.0
    out    = (8.6293e-6/np.sqrt(T))*factor*(1.602e-12*0.0079)*np.exp((-1.602e-12*0.0079)/(1.38065e-16*T))

    return out

def compute_lambda_CII_e(T):
    """
    Maxwellian-averaged collision rates with e-
    using expression from Goldsmith et al. 2012.
    see Appendix A, Ferrara et al. 2019
    """
    factor= (0.67*(T)**0.13)/2.0
    out   = (8.6293e-6/np.sqrt(T))*factor*(1.602e-12*0.0079)*np.exp((-1.602e-12*0.0079)/(1.38065e-16*T))

    return out


def compute_lambda_CIII_e(T):
    """
    Maxwellian-averaged collision rates with e-
    using expression from Appendix A, Ferrara et al. 2019
    """
    factor=1.265-0.0174e-4*T
    out = (8.6293e-6/np.sqrt(T))*factor*(1.602e-12*6.54)*np.exp((-1.602e-12*6.54)/(1.38065e-16*T))
    
    return out







import numpy as np
import scipy
import pyneb as pn
from astroglam.ion_structure import compute_Ni,compute_NF,compute_NHIyi,compute_NHIy0
from astroglam.cooling_rates import compute_lambda_CII_h,compute_lambda_CII_e, compute_lambda_CIII_e
from astroglam.atomic_data import *
from astroglam.empirical import *

O3 = pn.Atom('O', 3) #load the OIII ion from pyneb

def oxygen_abundance(Z, usrdefn=0):
    """
    Set the oxygen abundance. 
    We assume Asplund et al 2009 abundance at Zsun and that Ao scales linearly with Z. Z in solar units
    """
    if(usrdefn==0):
        Ao = 4.90e-4
    else:
        Ao = usrdefn
    return Ao*Z

def carbon_abundance(Z, usrdefn=0):
    """
    Set the carbon abundance. 
    We assume Asplund et al 2009 abundance at Zsun and that Ac scales linearly with Z. Z in solar units
    """
    if(usrdefn==0):
        Ac = 2.7e-4 
    else:
        Ac = usrdefn
    return Ac*Z
   
def compute_sigmag(Sigma_sfr, k, ks_fit=1.4):
    """
    Gas surface density from the SFR surface density assuming modified Kennicutt-Schmidt relation. 
    Sigma_sfr in Msun/yr/kpc^2
    k is the burstiness parameter
    """
    out=(((k*(10.0**-12))**-1)*Sigma_sfr)**(1./ks_fit)
    return out

def compute_U(Sigma_sfr, Sigma_g):
    """
    Ionizaton parameter as a function of star formation and gas surface densities. 
    reference:
      Eq. 38 Ferrara et al. 2019
    inputs:
      Sigma_sfr in Msun/yr/kpc^2
      Sigma_g   in Msun/kpc^2
    """
    out= (1.7e+14)*(Sigma_sfr/(Sigma_g*Sigma_g))
    return out

def compute_U_and_N(Z,k,Sigma_sfr,ks_fit=1.4 ):

    """
    wrapper to compute ionization parameter and column density
    """

    # the gas surface density is computed from the modified SK
    surface_density_gas = compute_sigmag(Sigma_sfr=Sigma_sfr, k=k,ks_fit=ks_fit)
    # convert the gas surface density to a column density
    column_density      = (surface_density_gas*10**22.0)/7.5e+7
    # compute the ionization parameter
    ionization_parameter= compute_U(Sigma_sfr=Sigma_sfr, Sigma_g = surface_density_gas)

    return ionization_parameter,column_density 


def compute_flux_cii_density_bound(n, Z, U, column,  TPDR=100.0, THII=1.e+4):
    """
    Emerging [CII] flux for the density bounded case
    Eq. 34 in Ferrara et al. 2019
    """

    # column density in the density bound case
    N_HIy0 = compute_NHIy0(U, Z,column)

    column_ionized = N_HIy0

    # compute the cooling contributions with and without LTE approximation
    if(n<=n_crit_CII):
        cooling_ionized    = n*compute_lambda_CII_e(T=THII)
    else:
        cooling_ionized    = (g2_cii/g1_cii)*pop_LTE(T_in=THII,E_in=E12_158um)*A21_158um*(ev2erg*E12_158um)

    out = carbon_abundance(Z, usrdefn=2.51e-4)*(column_ionized*cooling_ionized)

    return out

def pop_LTE(T_in,E_in):

    """
    return the local thermodynamical equilibrium population
    convenient conversion for
      E_in is the level energy difference in eV
      T the temperature of the gas in K
    """

    out = np.exp((-ev2erg*E_in)/(1.38065e-16*T_in))

    return out

def compute_flux_cii_ionization_bound_N0(n, Z, U, column, TPDR=100.0, THII=1.e+4):

    """
    Emerging [CII] flux in the ionization bounded case, and for NF> N0
    Equations 30, 31 in Ferrara et al. 2019
    """

    # ionized column density
    N_i= compute_Ni(U,Z)
    # neutral column density in the ionized layer
    N_HIyi = compute_NHIyi(U, Z)

    column_neutral = column - N_i
    column_ionized = N_HIyi

    # compute the cooling contributions with and without LTE approximation
    if(n<=n_crit_CII):
        cooling_neutral    = n*compute_lambda_CII_h(T=TPDR)
        cooling_ionized    = n*compute_lambda_CII_e(T=THII)
    else:
        cooling_neutral    = (g2_cii/g1_cii)*pop_LTE(T_in=TPDR,E_in=E12_158um)*A21_158um*(ev2erg*E12_158um)
        cooling_ionized    = (g2_cii/g1_cii)*pop_LTE(T_in=THII,E_in=E12_158um)*A21_158um*(ev2erg*E12_158um)

    out = carbon_abundance(Z, usrdefn=2.51e-4)*( column_neutral*cooling_neutral + column_ionized*cooling_ionized)

    return out

def compute_flux_cii_ionization_bound_NF(n, Z, U, column, TPDR=100.0, THII=1.e+4):
    """
    Emerging [CII] flux in the ionization bounded case, and for NF< N0
    Equations 30, 32 in Ferrara et al. 2019
    """

    # ionized column density
    N_i    = compute_Ni(U,Z)
    # NL column density
    N_F    = compute_NF(U, Z)
    # neutral column density in the ionized layer
    N_HIyi = compute_NHIyi(U, Z)

    column_neutral = N_F - N_i
    column_ionized = N_HIyi

    # compute the cooling contributions with and without LTE approximation
    if(n<=n_crit_CII):
        cooling_neutral    = n*compute_lambda_CII_h(T=TPDR)
        cooling_ionized    = n*compute_lambda_CII_e(T=THII)
    else:
        cooling_neutral    = (g2_cii/g1_cii)*pop_LTE(T_in=TPDR,E_in=E12_158um)*A21_158um*(ev2erg*E12_158um)
        cooling_ionized    = (g2_cii/g1_cii)*pop_LTE(T_in=THII,E_in=E12_158um)*A21_158um*(ev2erg*E12_158um)

    out = carbon_abundance(Z, usrdefn=2.51e-4)*( column_neutral*cooling_neutral + column_ionized*cooling_ionized)
        
    return out

def Sigma_CII158(logn, Z, k, Sigma_sfr):

    """
    compute the CII surface brightenss
    Eq. 35 in Ferrara et al. 2019
    """

    UU , column_density = compute_U_and_N(Z=Z,k=k,Sigma_sfr=Sigma_sfr)
    N_i                 = compute_Ni(U=UU,Z=Z)
    N_F                 = compute_NF(U=UU,Z=Z)
    n                   = 10**logn

    if(column_density<N_i):
         flux = compute_flux_cii_density_bound(n = n, Z= Z, U = UU, column=column_density   , TPDR=100.0, THII=1e+4)
    elif(column_density<N_F):
         flux = compute_flux_cii_ionization_bound_N0(n = n, Z= Z, U = UU, column=column_density, TPDR=100.0, THII=1e+4)
    else:
         flux = compute_flux_cii_ionization_bound_NF(n = n, Z= Z, U = UU, column=column_density, TPDR=100.0, THII=1e+4)

    # flux to surface brightness conversion
    out = flux*2.474e+9

    return out

def compute_flux_ion(n, Z, U, THII,line="88um"):

    """
    flux for the line emission from the ionized gas, 
    Emissivity computed with Pyneb.
    details can be found in Vallini et al. 2021.
    """
    # ionized column density
    N_i           = compute_Ni(U=U,Z=Z)

    # get emissivity and relative population for the selected line: 
    em_ion = None
    ion_elem = None
   
    if line == "52um":
        em_ion        = O3.getEmissivity(THII, n, wave='51.8m') # erg s^-1 cm^3
        ion_elem ='o3'
    if line == "88um":
        em_ion        = O3.getEmissivity(THII, n, wave='88.3m') # erg s^-1 cm^3
        ion_elem ='o3'
    if line == "5007A":
        em_ion        = O3.getEmissivity(THII, n, wave='5007A') # erg s^-1 cm^3
        ion_elem ='o3'
    if line =='Halpha':
        em_ion = 0
        ion_elem='hii'

    
    #correction for the presence of OII in the ionized region
    if(ion_elem =='o3'):
        fo3           = np.array([0.10994503, 0.73298314, 0.96966708])
        Uo3           = np.array([-3.5, -2.5, -1.5])
        Xoiii         = np.interp(np.log10(U), Uo3, fo3)
        Nh_ion_elem   = oxygen_abundance(Z, usrdefn=3.19e-4)*Xoiii*N_i
    else:
        Nh_ion_elem    = N_i

    
    out           = em_ion * n * Nh_ion_elem

    return out

def compute_flux_ciii(n, Z, U, THII=1.e+4):
    """
    compute the CIII surface brightenss
    as outlined in Ferrara et al. 2019
    """
    
    # compute the cooling contributions with LTE approximation  
    cooling_ionized    = n*compute_lambda_CIII_e(T=THII)
    # ionized column density
    column_ionized= compute_Ni(U,Z)
    
    out = carbon_abundance(Z, usrdefn=2.51e-4)*(column_ionized*cooling_ionized)
    
    return out


def Sigma_CIII(logn, Z, k, Sigma_sfr):

    """
    compute the CIII surface brightenss
     in Ferrara et al. 2019
    """

    UU , __ = compute_U_and_N(Z=Z,k=k,Sigma_sfr=Sigma_sfr)
    n                   = 10**logn

    flux = compute_flux_ciii(n = n, Z= Z, U = UU, THII=1e+4)

    # flux to surface brightness conversion
    out = flux*2.474e+9

    return out

def Sigma_OIII88(logn, Z, k, Sigma_sfr):
    UU , __ =  compute_U_and_N(Z=Z,k=k,Sigma_sfr=Sigma_sfr)
    ff      = compute_flux_ion(n=10.0**logn, Z=Z, U=UU, THII=1e+4,line="88um")
    out     = ff*2.474e+9 # 
    return out

def Sigma_OIII52(logn, Z, k, Sigma_sfr):
    UU , __ =  compute_U_and_N(Z=Z,k=k,Sigma_sfr=Sigma_sfr)
    ff      = compute_flux_ion(n=10.0**logn, Z=Z, U=UU, THII=1e+4,line="52um")
    out     = ff*2.474e+9
    return out

def Sigma_OIII5007(logn, Z, k, Sigma_sfr):
    UU , __ =  compute_U_and_N(Z=Z,k=k,Sigma_sfr=Sigma_sfr)
    ff      = compute_flux_ion(n=10.0**logn, Z=Z, U=UU, THII=1e+4,line="5007A")
    out     = ff*2.474e+9
    return out

def Delta(logn, Z, k, Sigma_sfr):

    out = np.log10(Sigma_CII158(logn=logn, Z=Z, k=k, Sigma_sfr=Sigma_sfr))-np.log10(delooze_fit_resolved(Sigma_sfr=Sigma_sfr))
    return out







# glam: Galaxy Line Analyzer with MCMC
#Copyright (c) 2024 Livia Vallini

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

# IMPORT PACKAGES
import os
from astroglam.atomic_data import *
from astroglam.ion_structure import *
from astroglam.cooling_rates import *
from astroglam.empirical import *
from astroglam.emission_models import *
from astroglam.MCMC_routines import *

# Version of glam package
__version__='0.1.0'



from __future__ import print_function
import numpy as np
import os
from multiprocessing import Pool
import emcee
from astroglam.emission_models import Delta, Sigma_CII158, Sigma_OIII88, Sigma_OIII52, Sigma_OIII5007, Sigma_CIII
from astroglam.empirical import delooze_fit_resolved

class galaxy_template:

    def __init__(
        self,
        Sigma_SFR=2.0,  # Msun/yr/kpc
        Sigma_CII=3.0e7,  # Lsun/kpc^2
        Sigma_ion=7.0e7  # Lsun/kpc^2
        #
        ,
        rel_err_Sigma_CII=0.2,  # relative error on the SigmaCII
        rel_err_Sigma_ion=0.2,  # relative error on the SigmaOIII
        rel_err_Delta=0.2,  # relative error on the Delta
        ion_tracer='88um',
    ):

        # setup the template for the data
        #
        self.Sigma_SFR = None  # Msun/yr/kpc^2
        self.Sigma_CII = None  # Lsun/kpc^2
        self.Sigma_ion = None  # Lsun/kpc^2
        #
        self.rel_err_Sigma_CII = None
        self.rel_err_Sigma_ion = None
        self.rel_err_Delta = None
        
        self.ion_tracer= None

        # set the input errors
        self.set_relative_errors(
            rel_err_Sigma_CII=rel_err_Sigma_CII,
            rel_err_Sigma_ion=rel_err_Sigma_ion,
            rel_err_Delta=rel_err_Delta,
        )

        # set the input data
        self.set_data(Sigma_SFR=Sigma_SFR, Sigma_CII=Sigma_CII, Sigma_ion=Sigma_ion, ion_tracer=ion_tracer)

    def set_relative_errors(
        self, rel_err_Sigma_CII=None, rel_err_Sigma_ion=None, rel_err_Delta=None
    ):

        if rel_err_Sigma_CII is not None:
            self.rel_err_Sigma_CII = rel_err_Sigma_CII
        if rel_err_Sigma_ion is not None:
            self.rel_err_Sigma_ion = rel_err_Sigma_ion
        if rel_err_Delta is not None:
            self.rel_err_Delta = rel_err_Delta

    def set_data(self, Sigma_SFR=None, Sigma_CII=None, Sigma_ion=None, ion_tracer=None):

        if Sigma_SFR is not None:
            self.Sigma_SFR = Sigma_SFR
        if Sigma_CII is not None:
            self.Sigma_CII = Sigma_CII
        if Sigma_ion is not None:
            self.Sigma_ion = Sigma_ion
        if ion_tracer is not None:
            self.ion_tracer = ion_tracer

    def data_for_MCMC(self):
        y = np.array([self.Deltagalaxy(), self.Sigma_CII, self.Sigma_ion])
        yerr = np.array(
            [
                self.Deltagalaxy() * self.rel_err_Delta,
                self.Sigma_CII * self.rel_err_Sigma_CII,
                self.Sigma_ion * self.rel_err_Sigma_ion,
            ]
        )

        par = self.Sigma_SFR
        label=self.ion_tracer

        return y, yerr, par, label

    def Deltagalaxy(self):

        out = np.log10(self.Sigma_CII) - np.log10(delooze_fit_resolved(self.Sigma_SFR))

        return out

    def print_info(self):

        print("Galaxy input data")
        print("  ion tracer   = ", self.ion_tracer)
        #
        print("  Sigma_SFR         = ", self.Sigma_SFR, "Msun/yr/kpc^2")
        print("  Sigma_CII         = ", self.Sigma_CII, "Lsun/kpc^2")
        print("  Sigma_ion        = ", self.Sigma_ion, "Lsun/kpc^2")
        #
        print("  relative error on Delta       = ", 100.0 * self.rel_err_Delta, "%")
        print("  relative error on Sigma_CII  = ", 100.0 * self.rel_err_Sigma_CII, "%")
        print(
            "  relative error on Sigma_ion  = ", 100.0 * self.rel_err_Sigma_ion, "%"
        )


class MC_model:

    def __init__(
        self
        # priors
        ,
        lognMIN=0.5,
        lognMAX=3.5,  # minimum and maximum log density [cm-3]
        logkMIN=-1.0,
        logkMAX=2.5,  # minimum and maximum log k_s
        logZMIN=-1.5,
        logZMAX=0.0  # minimum and maximum log metallicity [Zsun]
        #
        # MCMC parameters
        ,
        n_walkers=10,
        steps=200,
        burn_in=50
        #
        # starting point for the walkers
        ,
        logn0=2.0,
        logZ0=-0.5,
        logk0=0.3
        # backend fname
        ,
        bkh_fname=None,
    ):

        # backend init
        self.bkh_fname = None

        # priors for the MCMC
        self.lognMIN = None
        self.lognMAX = None
        self.logkMIN = None
        self.logkMAX = None
        self.logZMIN = None
        self.logZMAX = None

        # walker init
        self.logn0 = None
        self.logZ0 = None
        self.logk0 = None

        # eemc parameters
        self.n_dim = 3
        self.n_walkers = None
        self.steps = None
        self.burn_in = None

        self.galaxy_data = None

        self.pool = None

        self.set_priors(
            lognMIN=lognMIN,
            lognMAX=lognMAX,
            logkMIN=logkMIN,
            logkMAX=logkMAX,
            logZMIN=logZMIN,
            logZMAX=logZMAX,
        )

        self.set_walkers(logn0=logn0, logZ0=logZ0, logk0=logk0)
        self.set_backend(bkh_fname=bkh_fname)
        self.set_mc_parameters(n_walkers=n_walkers, steps=steps, burn_in=burn_in)
       
        self.Y = None
        self.YERR = None
        self.SSFR = None
        self.LINELABEL = None

    def check_consistency(self):

        # check the priors
        ok_priors = self.lognMIN < self.lognMAX
        ok_priors = ok_priors and self.logkMIN < self.logkMAX
        ok_priors = ok_priors and self.logZMIN < self.logZMAX
        if not ok_priors:
            print("Priors look funky, i.e. min >= max for some of them")

        ok_init = self.check_bounds(theta=tuple([self.logn0, self.logZ0, self.logk0]))
        if not ok_init:
            print("Initial position of the walkers is out of the priors")

        ok_data = isinstance(self.galaxy_data, galaxy_template)
        if not ok_data:
            print("galaxy data is not a galaxy_template() class")

        NoneType = type(None)
        no_backend = isinstance(self.bkh_fname, NoneType)
        ok_backend = not (no_backend)
        if not ok_backend:
            print("You need to provide the fname for the backend, use set_backend()")

        return ok_priors and ok_init and ok_data and ok_backend

    def run_model_parallel(self, proc=None):

        ok = self.check_consistency()
        if not ok:
            print("Problem in the initialization, aborting")
        assert ok

        if(1):
            print("about to run (with parallelization) ...")
            self.print_info()
            print("galaxy data")
            self.galaxy_data.print_info()

        # get galaxy data in the format for the MCMC
        y, yerr, par,ll = self.galaxy_data.data_for_MCMC()

        # set them as global variables
        self.set_global(y, yerr, par)
        
        # set the line label as global param as well
        self.set_linelabel_global(ll)
        
        # init walkers
        starting_point = [self.logn0, self.logZ0, self.logk0]
        pos = [
            starting_point + 1e-5 * np.random.randn(self.n_dim)
            for i in range(self.n_walkers)
        ]

        # Set up the backend
        backend = emcee.backends.HDFBackend(self.bkh_fname)
        backend.reset(self.n_walkers, self.n_dim)

        if proc is None:
            with Pool() as pool:
                # set the emcee
                sampler = emcee.EnsembleSampler(
                    self.n_walkers, self.n_dim, self.lnprob_global, pool=pool
                )
                # run the thing
                sampler.run_mcmc(pos, self.steps, progress=True)
                # tau            = sampler.get_autocorr_time(quiet=True)
            flat_samples = sampler.get_chain(discard=self.burn_in, flat=True)
            return flat_samples

        else:
            with Pool(processes=proc) as pool:
                # set the emcee
                sampler = emcee.EnsembleSampler(
                    self.n_walkers, self.n_dim, self.lnprob_global, pool=pool
                )
                # run the thing
                sampler.run_mcmc(pos, self.steps, progress=True)
                # tau            = sampler.get_autocorr_time()
            flat_samples = sampler.get_chain(discard=self.burn_in, flat=True)
            return flat_samples

    def run_model(self, verbose=True):

        ok = self.check_consistency()
        if not ok:
            print("Problem in the initialization, aborting")
        assert ok

        if verbose:
            print("about to run (w/o parallelization)....")
            self.print_info()
            print("galaxy data")
            self.galaxy_data.print_info()

        # get galaxy data in the format for the MCMC
        y, yerr, par, ll = self.galaxy_data.data_for_MCMC()
        

        # init walkers
        starting_point = [self.logn0, self.logZ0, self.logk0]
        pos = [
            starting_point + 1e-5 * np.random.randn(self.n_dim)
            for i in range(self.n_walkers)
        ]

        # Set up the backend
        backend = emcee.backends.HDFBackend(self.bkh_fname)
        backend.reset(self.n_walkers, self.n_dim)

        # set the emcee
        sampler = emcee.EnsembleSampler(
            self.n_walkers,
            self.n_dim,
            self.lnprob,
            args=(y, yerr, par, ll),
            backend=backend,
        )
        # run the thing
        sampler.run_mcmc(pos, self.steps, progress=True)
        tau = sampler.get_autocorr_time(quiet=True)
        flat_samples = sampler.get_chain(discard=self.burn_in, flat=True)

        return flat_samples

    def set_mc_parameters(self, n_walkers=None, steps=None, burn_in=None):

        # set priors for the model
        # if input is None, no change is done
        if n_walkers is not None:
            self.n_walkers = n_walkers
        if steps is not None:
            self.steps = steps
        if burn_in is not None:
            self.burn_in = burn_in

    def set_backend(self, bkh_fname=None):
        # set the name for the bkh file
        # if input is None, no change is done
        if bkh_fname is not None:
            self.bkh_fname = bkh_fname  #

    def set_priors(
        self,
        lognMIN=None,
        lognMAX=None,
        logkMIN=None,
        logkMAX=None,
        logZMIN=None,
        logZMAX=None,
    ):

        # set priors for the model
        # if input is None, no change is done
        if lognMIN is not None:
            self.lognMIN = lognMIN  #
        if lognMIN is not None:
            self.lognMAX = lognMAX  # maximum log density [cm-3]
        if logkMIN is not None:
            self.logkMIN = logkMIN  # minimum log k_s
        if logkMAX is not None:
            self.logkMAX = logkMAX  # maximum log k_s
        if logZMIN is not None:
            self.logZMIN = logZMIN  #
        if logZMAX is not None:
            self.logZMAX = logZMAX  # maximum log metallicity [Zsun]

    def set_galaxy_data(self, galaxy_data=None):
        self.galaxy_data = galaxy_data

    def set_walkers(self, logn0=None, logZ0=None, logk0=None):

        if logn0 is not None:
            self.logn0 = logn0
        if logZ0 is not None:
            self.logZ0 = logZ0
        if logk0 is not None:
            self.logk0 = logk0

    def check_bounds(self, theta):
        logn, logZ, logk = theta

        ok = self.lognMIN < logn < self.lognMAX
        ok = ok and self.logkMIN < logk < self.logkMAX
        ok = ok and self.logZMIN < logZ < self.logZMAX

        return ok

    def model(self, theta, ssfr, ll):

        logn, logZ, logk = theta

        Z, k = 10**logZ, 10**logk

        delta_galaxy = Delta(logn=logn, Z=Z, k=k, Sigma_sfr=ssfr)
        sigma_cii_galaxy = Sigma_CII158(logn=logn, Z=Z, k=k, Sigma_sfr=ssfr)
        
        if(ll=='88um'):
            sigma_ion_galaxy = Sigma_OIII88(logn=logn, Z=Z, k=k, Sigma_sfr=ssfr)
        if(ll=='52um'):
            sigma_ion_galaxy = Sigma_OIII52(logn=logn, Z=Z, k=k, Sigma_sfr=ssfr)
        if(ll=='OIII5007'):
            sigma_ion_galaxy = Sigma_OIII5007(logn=logn, Z=Z, k=k, Sigma_sfr=ssfr)
        if(ll=='CIII'):
            sigma_ion_galaxy = Sigma_CIII(logn=logn, Z=Z, k=k, Sigma_sfr=ssfr)
            
        return delta_galaxy, sigma_cii_galaxy, sigma_ion_galaxy

    def lnlike(self, theta, y, yerr, ssfr, ll):
        mod = np.asarray(self.model(theta=theta, ssfr=ssfr, ll=ll))
        inv_sigma2 = 1.0 / (yerr**2)
        out = -0.5 * (np.sum((y - mod) ** 2 * inv_sigma2))
        return out

    def lnprob(self, theta, y, yerr, ssfr, ll):
        lp = self.lnprior(theta)

        if not np.isfinite(lp):
            out = -np.inf
        else:
            out = lp + self.lnlike(theta, y, yerr, ssfr, ll)

        return out
    
    def lnprior(self,theta):
       #flat priors on logn, logk, and log Z

        ok = self.check_bounds(theta=theta)

        if  ok:
            out = 0.0
        else:
            out = -np.inf
       
        return out
    
    def set_global(self, a, b, c):
        
        self.Y=a
        self.YERR=b
        self.SSFR=c
        
    def set_linelabel_global(self, ll):
        
        self.LINELABEL=ll
        
    
    def model_global(self,theta):

        logn, logZ, logk = theta

        Z, k = 10**logZ, 10**logk

        delta_galaxy = Delta(logn=logn, Z=Z, k=k, Sigma_sfr=self.SSFR)
        sigma_cii_galaxy = Sigma_CII158(logn=logn, Z=Z, k=k, Sigma_sfr=self.SSFR)
        
        if(self.LINELABEL=='88um'):
            sigma_ion_galaxy = Sigma_OIII88(logn=logn, Z=Z, k=k, Sigma_sfr=self.SSFR)
        if(self.LINELABEL=='52um'):
            sigma_ion_galaxy = Sigma_OIII52(logn=logn, Z=Z, k=k, Sigma_sfr=self.SSFR)
        if(self.LINELABEL=='OIII5007'):
            sigma_ion_galaxy = Sigma_OIII5007(logn=logn, Z=Z, k=k, Sigma_sfr=self.SSFR)
        if(self.LINELABEL=='CIII'):
            sigma_ion_galaxy = Sigma_CIII(logn=logn, Z=Z, k=k, Sigma_sfr=self.SSFR)
            
        return delta_galaxy, sigma_cii_galaxy, sigma_ion_galaxy


    def lnlike_global(self,theta):
        mod              = np.asarray(self.model_global(theta=theta))
        inv_sigma2       = 1.0/(self.YERR**2)
        out              = -0.5*(np.sum((self.Y-mod)**2*inv_sigma2))
        return out
   
    
    def lnprob_global(self,theta):
        lp = self.lnprior(theta)

        if not np.isfinite(lp):
            out = -np.inf
        else:
            out = lp + self.lnlike_global(theta)

        return out
    

    def print_info(self):
        print("MC parameters")
        print("  n_walkers", self.n_walkers)
        print("  steps    ", self.steps)
        print("  burn_in  ", self.burn_in)

        print("Priors")
        print(
            "{:10} {:10} {:10}".format(self.lognMIN, "< log(n/cm^-3) <", self.lognMAX)
        )
        print(
            "{:10} {:10} {:10}".format(self.logkMIN, "< log(k_s)     <", self.logkMAX)
        )
        print(
            "{:10} {:10} {:10}".format(self.logZMIN, "< log(Z/Z_sun) <", self.logZMAX)
        )

        print("walkers starting point")
        print("  log n ", self.logn0)
        print("  log Z ", self.logZ0)
        print("  log k ", self.logk0)

        print("Backend fname")
        print("  fname ", self.bkh_fname)
        
        







# glam: Galaxy Line Analyzer with MCMC
#Copyright (c) 2024 Livia Vallini

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
import numpy as np

"""
  this module contains the equations to compute the ionization structure
  inputs:
    U is the ionzation parameter
    Z is the metallicity in Zsun
  outputs:
    column densities, measured in cm^-2
"""

def compute_Ni(U,Z):
    """
    Column density of the ionized gas
    Eq. 14 in Ferrara et al. 2019. 
    """
    ND     = compute_Nd(Z)
    tau_sd = (1e+23*U)/ND
    NN     = ND*np.log((1+tau_sd)/(1+(tau_sd/np.exp(1.0))))
    return NN

def compute_Nd(Z):
    """
    Column density corresponding to A_V = 1 
    Eq. 9c in Ferrara et al. 2019
    """
    Ndsolar=1.7e21 #cm^-2
    return Ndsolar/Z

def compute_chi_of_U(Z):
    """
    Eq. 22 in Ferrara et al. 2019
    """
    
    chi = 8.7e+4*Z

    return chi

def compute_chi_prime(U, Z):
    """
    Eq. 22b in Ferrara et al. 2019
    """
    ww       = compute_w_of_D(Z)
    chi      = compute_chi_of_U(U)
    chiprime = ww*chi
    return chiprime

def compute_w_of_D(Z):
    """
    Factor related to the abs. of LW photons
    see Eq. 24 in Ferrara et al. 2019 (see also Sternberg et al. 2014)
    """
    w = 1.0/(1.0+ 0.9*(Z)**0.5)
    return w

def compute_NF(U,Z):
    """
    Column density at which the Lyman-Werner flux vanishes
    Eq. 28 Ferrara et al. 2019
    """
    ND  = compute_Nd(Z)
    out = ND*np.log(1+compute_chi_prime(U,Z))
    return out

def compute_NHIyi(U, Z):
    """
    HI column in the ionized layer
    Eq. 13 and 14 in Ferrara et al. 2019
    """
    Ns     = 1e+23*U
    tau_s  = 2.7e+5 * U
    out    = (Ns/tau_s) * (1.0 - compute_Ni(U,Z)/compute_Nd(Z))
    return out

def compute_NHIy0(U, Z, N0):
    """
    Ionized column density in the case of the density bounded regime
    Eq. 33 in Ferrara et al. 2019
    """
    Ns      = 1e+23*U
    y0      = N0/Ns
    tau_sd  = 59.0 * U * Z
    tau_s   = 2.7e+5 * U
    out     = (Ns/tau_s) * np.log(tau_sd/np.abs((np.exp(tau_sd*y0)-tau_sd-1.0)))
    return out





# glam: Galaxy Line Analyzer with MCMC
#Copyright (c) 2024 Livia Vallini

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
import numpy as np

def delooze_fit(logsfr):
    return (logsfr + 7.06)/1.0

def delooze_fit_resolved(Sigma_sfr):
    logSigma_cii=(np.log10(Sigma_sfr) +6.99)/0.93
    return 10**logSigma_cii

def delooze_delta(Sigma_sfr,Sigma_cii):
    return np.log10(Sigma_cii) - np.log10(delooze_fit_resolved(Sigma_sfr))






from setuptools import setup, find_packages
import pathlib
here = pathlib.Path(__file__).parent.resolve()
# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")
# Setup
setup(
 name="astroglam",
 version="0.1.0",
 description="A code to leverage ALMA and JWST data to infer the gas density, metallicity, and burstiness of galaxies.",
 long_description=long_description,
 long_description_content_type="text/markdown",
 url="https://github.com/lvallini/",
 author="Livia Vallini",
 author_email="livia.vallini@inaf.it",
 classifiers=[
"Programming Language :: Python :: 3",
"License :: OSI Approved :: MIT License", 
"Operating System :: OS Independent",
"Topic :: Scientific/Engineering :: Astronomy",
"Intended Audience :: Science/Research"
 ],
 package_dir={"": "src"},
 packages=find_packages(where="src"),
 include_package_data=True,
 package_data={
 "": ["resources/*.csv"],
 "": ["docs/*.ipynb"],
 },
 python_requires=">=3.9, <4",
 install_requires=[
  'contourpy>=1.2.0',
  'corner>=2.2.2',
  'cycler>=0.11.0',
  'emcee>=3.1.6',
  'fonttools>=4.51.0',
  'h5py>=3.11.0',
  'kiwisolver>=1.4.4',
  'matplotlib>=3.9.2',
  'numpy<2.0',
  'packaging==24.1',
  'pillow>=10.4.0',
  'pip>=24.2',
  'PyNeb>=1.1.16',
  'pyparsing>=3.1.2',
  'python-dateutil>=2.9.0.post0',
  'scipy>=1.13.1',
  'setuptools>=59',
  'six>=1.16.0',
  'tornado>=6.4.1',
  'unicodedata2>=15.1.0',
  'wheel>=0.44.0',
  "tqdm>=4.66.5",
],
 project_urls={ # Optional

 "Bug Reports": "https://github.com/lvallini/MCMC_galaxyline_analyzer",
 "Source": "https://github.com/lvallini/MCMC_galaxyline_analyzer",
 },
)



import numpy as np
import astroglam as gl
import corner
import os

if __name__ == '__main__':
       # Instantiate your class and run the method
   

	mcr = gl.MC_model()
	mcr.set_priors(lognMIN = 0.0, lognMAX= 4.5,logZMIN = -2, logZMAX= 0.1, logkMIN = -0.5, logkMAX= 3.5)

	mcr.set_walkers(logn0=1.0,logZ0=-0.66,logk0 = 0.92)
	mcr.set_mc_parameters(n_walkers = 10, steps = 10000, burn_in   = 100)

	galaxy_exampleOIII = gl.galaxy_template()

	# load the cii, ciii, oiii, sfr in the cell

	Sigma_CIIcell=10**7
	Sigma_ioncell=10**7
	Sigma_SFRcell=10.0
	print('THIS IS A TEST RUN: ....')
            
	# run mcmc for oiii
	filename_backendoiii='oiii.h5'
	galaxy_exampleOIII.set_data(Sigma_SFR=Sigma_SFRcell, Sigma_CII=Sigma_CIIcell, Sigma_ion=Sigma_ioncell, ion_tracer='OIII5007')
	galaxy_exampleOIII.set_relative_errors(rel_err_Sigma_CII=0.2, rel_err_Sigma_ion=0.2, rel_err_Delta=0.2)
	mcr.set_backend(bkh_fname=filename_backendoiii)
	mcr.set_galaxy_data(galaxy_data = galaxy_exampleOIII)
	flat_samplesOIII = mcr.run_model()
            
	fig = corner.corner(flat_samplesOIII, labels=["log(n)", r"log(Z/Z$_{\odot}$)", r"log($\kappa_s$)"])

	fig.savefig("test.png")

	print('EXITED OK! GLAM SHOULD WORK FINE')
